import logging
import random
from typing import Optional, Union

import numpy as np
from ai2thor.controller import Controller
from attr import Attribute, field
from attrs import define
from houses.constants import PROCTHOR_INITIALIZATION
from houses.utils.types import SamplingVars, Split

from .ceiling_height import sample_ceiling_height
from .color_objects import randomize_object_color
from .doors import add_doors
from .exterior_walls import add_exterior_walls
from .floorplans import consolidate_walls, generate_floorplan
from .generation import (
    create_house,
    find_walls,
    get_floor_polygons,
    get_xz_poly_map,
    scale_boundary_groups,
)
from .house import House
from .interior_boundaries import sample_interior_boundary
from .layer import assign_layer_to_rooms
from .lights import add_lights
from .materials import randomize_materials
from .object_states import randomize_states
from .objects import add_objects
from .room_specs import RoomSpec, room_spec_sampler
from .skyboxes import add_skybox
from .small_objects import add_small_objects
from .wall_objects import add_wall_objects


def timeout(func, args=(), kwargs={}, timeout_duration=1, default=None):
    import signal

    class TimeoutError(Exception):
        pass

    def handler(signum, frame):
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout_duration)
    try:
        result = func(*args, **kwargs)
    except TimeoutError as exc:
        result = default
        raise TimeoutError("Function timed out!")
    finally:
        signal.alarm(0)

    return result


@define
class HouseGenerator:
    split: Split = field()
    seed: Optional[int] = None
    room_spec: Optional[Union[RoomSpec, str]] = None
    interior_boundary: Optional[np.array] = None
    controller: Optional[Controller] = None

    @split.validator
    def valid_split(self, attribute: Attribute, value: Split) -> None:
        if value not in {"train", "val", "test"}:
            raise ValueError(f'split={value} must be in {{"train", "val", "test"}}')

    def __attrs_post_init__(self) -> None:
        if self.seed is None:
            self.seed = random.randint(0, 2 ** 15)
            logging.debug(f"Using seed {self.seed}")
        if self.seed is not None:
            self.set_seed(self.seed)
        if isinstance(self.room_spec, str):
            self.room_spec = room_spec_sampler[self.room_spec]

    def set_seed(self, seed: int) -> None:
        # TODO: Probably should not be done on a global level
        self.seed = seed
        random.seed(seed)
        np.random.seed(seed)
        if self.controller is not None:
            self.controller.step(action="SetRandomSeed", seed=seed, renderImage=False)

    def sample(self):
        """Sample a house specification compatible with AI2-THOR."""
        sampling_vars = SamplingVars.sample()

        # NOTE: preprocessing of the room_spec
        room_spec = (
            self.room_spec if self.room_spec is not None else room_spec_sampler.sample()
        )
        room_ids = set(room_spec.room_type_map.keys())

        if self.controller is None:
            # NOTE: assumes images are never used by this Controller.
            self.controller = Controller(quality="Low", **PROCTHOR_INITIALIZATION)
            if self.seed is not None:
                self.controller.step(
                    action="SetRandomSeed", seed=self.seed, renderImage=False
                )

        # NOTE: sample interior boundary
        for i in range(10):

            def build_floorplan():
                interior_boundary = (
                    self.interior_boundary
                    if self.interior_boundary is not None
                    else sample_interior_boundary(
                        num_rooms=len(room_ids),
                        dims=None if room_spec.dims is None else room_spec.dims(),
                    )
                )
                floorplan = generate_floorplan(
                    interior_boundary=interior_boundary, room_spec=room_spec
                )
                walls = find_walls(floorplan=floorplan)
                boundary_groups = consolidate_walls(walls=walls)
                boundary_groups = scale_boundary_groups(
                    boundary_groups=boundary_groups,
                    scale=sampling_vars.interior_boundary_scale,
                )
                xz_poly_map = get_xz_poly_map(
                    boundary_groups=boundary_groups, room_ids=room_ids
                )
                return interior_boundary, floorplan, walls, boundary_groups, xz_poly_map

            # NOTE: build floorplan
            try:
                (
                    interior_boundary,
                    floorplan,
                    walls,
                    boundary_groups,
                    xz_poly_map,
                ) = timeout(build_floorplan, timeout_duration=10)
                break
            except Exception as e:
                print(f"Failed to generate floorplan {i}!! {e}")
                continue
        else:
            print("Failed to generate interior boundary!!")
            raise Exception("Failed to generate interior boundary!!")

        ceiling_height = sample_ceiling_height()

        created_house = create_house(
            xz_poly_map=xz_poly_map,
            room_type_map=room_spec.room_type_map,
            ceiling_height=ceiling_height,
        )
        house_data = created_house["house"]
        wall_map = created_house["wallMap"]

        floor_polygons = get_floor_polygons(xz_poly_map=xz_poly_map)

        # NOTE: procedural placement
        door_polygons = add_doors(
            house=house_data,
            boundary_groups=boundary_groups,
            room_spec=room_spec,
            split=self.split,
            wall_map=wall_map,
        )
        randomize_materials(house=house_data)
        add_lights(
            house=house_data,
            floor_polygons=floor_polygons,
            ceiling_height=ceiling_height,
        )
        add_skybox(house=house_data)

        # NOTE: added after randomize materials on purpose
        add_exterior_walls(
            house=house_data,
            boundary_groups=boundary_groups,
        )

        rooms = add_objects(
            house=house_data,
            floor_polygons=floor_polygons,
            room_type_map=room_spec.room_type_map,
            split=self.split,
            door_polygons=door_polygons,
            controller=self.controller,
            max_floor_objects=sampling_vars.max_floor_objects,
        )

        add_wall_objects(
            house=house_data,
            rooms=rooms,
            boundary_groups=boundary_groups,
            split=self.split,
            room_type_map=room_spec.room_type_map,
            wall_map=wall_map,
            ceiling_height=ceiling_height,
        )

        add_small_objects(
            house=house_data,
            rooms=rooms,
            split=self.split,
            controller=self.controller,
        )

        randomize_object_color(house=house_data)
        randomize_states(house=house_data)

        assign_layer_to_rooms(house=house_data)

        return House(
            data=house_data,
            rooms=rooms,
            interior_boundary=interior_boundary,
            room_spec=room_spec,
        )
