"""Floor plan generation scripts.

This prototype reproduces `A Constrained Growth Method for Procedural Floor Plan
Generation <https://api.semanticscholar.org/CorpusID:18711477>`_ with some modifications
adapted from this [re-implementation](https://github.com/lucafons/procedural-city).
"""

import copy
from collections import defaultdict

import numpy as np
from houses.constants import OUTDOOR_ROOM_ID
from houses.utils.types import LeafRoom, MetaRoom

from .room_specs import RoomSpec

np.seterr(divide="raise", invalid="raise")


def consolidate_walls(walls):
    """Join neighboring walls together.

    Example: if one of the walls is ::

        [
            ((0, 0), (0, 1)),
            ((0, 1), (0, 2)),
            ((0, 2), (0, 3)),
            ((0, 3), (0, 4)),
            ((0, 4), (0, 5)),
            ((0, 5), (0, 6)),
            ((0, 6), (0, 7)),
            ((0, 7), (0, 8)),
            ((0, 8), (0, 9)),
            ((0, 0), (1, 0)),
            ((1, 0), (2, 0)),
            ((2, 0), (3, 0)),
            ((3, 0), (4, 0)),
            ((4, 0), (5, 0)),
            ((5, 0), (6, 0)),
            ((6, 0), (7, 0)),
            ((7, 0), (8, 0)),
            ((8, 0), (9, 0))
        ]

    it becomes ::

        {
            ((0, 0), (0, 9)),
            ((0, 0), (9, 0))
        }
    """
    out = dict()
    for wall_group_id, wall_pairs in walls.items():
        wall_map = defaultdict(lambda: set())
        wall_map = dict()
        for wall in wall_pairs:
            if wall[0] not in wall_map:
                wall_map[wall[0]] = set()
            wall_map[wall[0]].add(wall[1])

        did_update = True
        while did_update:
            did_update = False
            for w1_1 in wall_map.copy():
                if w1_1 not in wall_map:
                    continue
                break_outer = False
                for w1_2 in wall_map[w1_1]:
                    if w1_2 in wall_map:
                        w2_1 = w1_2
                        for w2_2 in wall_map[w2_1]:
                            if (
                                w1_1[0] == w1_2[0] == w2_1[0] == w2_2[0]
                                or w1_1[1] == w1_2[1] == w2_1[1] == w2_2[1]
                            ):
                                wall_map[w2_1].remove(w2_2)
                                if not wall_map[w2_1]:
                                    del wall_map[w2_1]

                                wall_map[w1_1].remove(w2_1)
                                wall_map[w1_1].add(w2_2)

                                did_update = True
                                break_outer = True
                                break
                        if break_outer:
                            break
                    if break_outer:
                        break
        out[wall_group_id] = set([(w1, w2) for w1 in wall_map for w2 in wall_map[w1]])
    return out


def generate_floorplan(interior_boundary, room_spec: RoomSpec):
    floor = interior_boundary.copy()

    # add 1-padding to the floor
    s1, s2 = floor.shape
    padded_floor = np.ones((s1 + 2, s2 + 2))
    padded_floor[1:-1, 1:-1] = floor

    recursive_plan(padded_floor, room_spec.spec)
    return padded_floor


def initial_placement(floor, rooms):
    weights = np.ones_like(floor)
    weights[floor == OUTDOOR_ROOM_ID] = 0
    for room in rooms:
        if np.sum(weights) == 0:
            raise Exception("Divide by 0!!")
        probs = weights.reshape((-1)) / np.sum(weights)
        ind = np.random.choice(len(weights.reshape((-1))), p=probs)

        # sets p=0 for immediate neighbors
        weights[
            max(0, int(ind / floor.shape[1]) - 1) : min(
                floor.shape[0], int(ind / floor.shape[1]) + 2
            ),
            max(0, ind % floor.shape[1] - 1) : min(
                floor.shape[0], ind % floor.shape[1] + 2
            ),
        ] = 0
        floor[int(ind / floor.shape[1]), ind % floor.shape[1]] = room.room_id
        room.x = int(ind / floor.shape[1])
        room.y = ind % floor.shape[1]
        room.w = 1
        room.h = 1


def select_room(rooms):
    ratios = np.array([room.r / (room.w * room.h) for room in rooms])
    return np.random.choice(len(rooms), p=(ratios / np.sum(ratios)))


def grow_rect(room, floor):
    f, b, r, l = 0, 0, 0, 0
    if (
        not (
            floor[
                min(floor.shape[0], room.x + room.w) : min(
                    floor.shape[0], room.x + 1 + room.w
                ),
                room.y : room.y + room.h,
            ]
            != 0
        ).any()
        and room.x + room.w != floor.shape[0]
    ):
        r = room.h
    if (
        not (floor[max(0, room.x - 1) : room.x, room.y : room.y + room.h] != 0).any()
        and room.x != 0
    ):
        l = room.h
    if (
        not (
            floor[
                room.x : room.x + room.w,
                min(floor.shape[1], room.y + room.h) : min(
                    floor.shape[1], room.y + 1 + room.h
                ),
            ]
            != 0
        ).any()
        and room.y + room.h != floor.shape[1]
    ):
        f = room.w
    if (
        not (floor[room.x : room.x + room.w, max(0, room.y - 1) : room.y] != 0).any()
        and room.y != 0
    ):
        b = room.w
    probs = np.array([r, l, f, b])
    if room.w * room.h >= room.r * 4:
        return False
    if all(probs == 0):
        return False
    choice = np.random.choice(4, p=(probs / np.sum(probs)))
    if choice == 0:
        floor[
            min(floor.shape[0], room.x + room.w) : min(
                floor.shape[0], room.x + 1 + room.w
            ),
            room.y : room.y + room.h,
        ] = room.room_id
        room.w += 1
        return True
    elif choice == 1:
        floor[max(0, room.x - 1) : room.x, room.y : room.y + room.h] = room.room_id
        room.x -= 1
        room.w += 1
        return True
    elif choice == 2:
        floor[
            room.x : room.x + room.w,
            min(floor.shape[1], room.y + room.h) : min(
                floor.shape[1], room.y + 1 + room.h
            ),
        ] = room.room_id
        room.h += 1
        return True
    else:
        floor[room.x : room.x + room.w, max(0, room.y - 1) : room.y] = room.room_id
        room.y -= 1
        room.h += 1
        return True


def grow_lshape(room, floor):
    f, b, r, l = 0, 0, 0, 0
    rb = []
    r = 0
    for i in range(room.h):
        if (
            floor[room.x + room.w, room.y + i] == 0
            and floor[room.x + room.w - 1, room.y + i] == room.room_id
        ):
            r += 1
            rb.append([room.x + room.w, room.y + i])
    lb = []
    l = 0
    for i in range(room.h):
        if (
            floor[room.x - 1, room.y + i] == 0
            and floor[room.x, room.y + i] == room.room_id
        ):
            l += 1
            lb.append([room.x - 1, room.y + i])
    fb = []
    f = 0
    for i in range(room.w):
        if (
            floor[room.x + i, room.y + room.h] == 0
            and floor[room.x + i, room.y + room.h - 1] == room.room_id
        ):
            f += 1
            fb.append([room.x + i, room.y + room.h])
    bb = []
    b = 0
    for i in range(room.w):
        if (
            floor[room.x + i, room.y - 1] == 0
            and floor[room.x + i, room.y] == room.room_id
        ):
            b += 1
            bb.append([room.x + i, room.y - 1])
    probs = np.array([r, l, f, b])
    if all(probs == 0):
        return False
    choice = np.argmax(probs)
    if choice == 0:
        for ind in rb:
            floor[ind[0], ind[1]] = room.room_id
        room.w += 1
        return True
    elif choice == 1:
        for ind in lb:
            floor[ind[0], ind[1]] = room.room_id
        room.x -= 1
        room.w += 1
        return True
    elif choice == 2:
        for ind in fb:
            floor[ind[0], ind[1]] = room.room_id
        room.h += 1
        return True
    else:
        floor[room.x : room.x + room.w, max(0, room.y - 1) : room.y][
            floor[room.x : room.x + room.w, max(0, room.y - 1) : room.y] == 0
        ] = room.room_id
        for ind in bb:
            floor[ind[0], ind[1]] = room.room_id
        room.y -= 1
        room.h += 1
        return True


def build_plan(floor, rooms):
    initial_placement(floor, rooms)
    grow_rooms = rooms.copy()
    while grow_rooms:
        curroom = select_room(grow_rooms)
        cangrow = grow_rect(grow_rooms[curroom], floor)
        if not cangrow:
            grow_rooms.remove(grow_rooms[curroom])
    grow_rooms = rooms.copy()
    while grow_rooms:
        curroom = select_room(grow_rooms)
        cangrow = grow_lshape(grow_rooms[curroom], floor)
        if not cangrow:
            grow_rooms.remove(grow_rooms[curroom])


def perimeters(floor, rooms):
    def find_walls_p(rooms, floor):
        """
        out[2] is the length of the wall.
        out[3] = bool True if vertical wal, False otherwise.
        """
        walls = []
        for room in rooms:
            if isinstance(room, MetaRoom):
                walls += find_walls_p(room.children, floor)
            else:
                x, y = find_start(room, floor)
                walls += crawl_rec(x, y, room.room_id, floor, 0, x, y)
        return list(dict.fromkeys(walls))

    walls = find_walls_p(rooms, floor)
    totalp = sum([wall[2] for wall in walls])
    return totalp


def evaluate_sizes(floor, rooms):
    totaldiff = 0
    totalr = np.sum([room.r for room in rooms])
    for room in rooms:
        size_r = np.count_nonzero(floor == room.room_id) / np.count_nonzero(floor != 1)
        target_r = room.r / totalr
        totaldiff += np.square((size_r - target_r))
    return totaldiff


def best_plan(floor, rooms):
    n = 100
    floor_arrs = [np.copy(floor) for _ in range(n)]
    room_arrs = [copy.deepcopy(rooms) for _ in range(n)]
    scores = []
    for i in range(n):
        donebuilding = False
        while not donebuilding:
            build_plan(floor_arrs[i], room_arrs[i])
            if np.count_nonzero(floor_arrs[i] == 0) > 0 or any(
                room.room_id not in np.unique(floor_arrs[i]) for room in rooms
            ):
                floor_arrs[i] = np.copy(floor)
                room_arrs[i] = copy.deepcopy(rooms)
            else:
                donebuilding = True
        scores.append(perimeters(floor_arrs[i], room_arrs[i]))
    sort = np.argsort(scores)
    n2 = 10
    shaped_floors = np.array(floor_arrs)[sort[:n2]]
    shaped_rooms = np.array(room_arrs)[sort[:n2]]
    scores = []
    for i in range(n2):
        scores.append(evaluate_sizes(shaped_floors[i], shaped_rooms[i]))
    best = np.argmin(scores)
    floor[:] = shaped_floors[best]
    rooms = shaped_rooms[best]


def recursive_plan(floor, rooms):
    cur_rooms = []
    for room in rooms:
        if isinstance(room, MetaRoom):
            cur_rooms.append(
                LeafRoom(room_id=room.room_id, ratio=room.r, room_type=room.room_type)
            )
        else:
            cur_rooms.append(room)
    curfloor = np.copy(floor)
    best_plan(curfloor, cur_rooms)
    for room in rooms:
        if isinstance(room, MetaRoom):
            levelfloor = np.copy(curfloor)
            levelfloor[levelfloor != room.room_id] = 1
            levelfloor[levelfloor == room.room_id] = 0
            levelrooms = room.children
            recursive_plan(levelfloor, levelrooms)
            curfloor[curfloor == room.room_id] = levelfloor[curfloor == room.room_id]
    floor[:] = curfloor


def crawl_rec(x, y, room_id, floor, dir, ox, oy):
    curx = x
    cury = y
    curlen = 0
    walls = []
    if dir == 0:
        while (
            floor[curx, cury] == room_id
            and floor[curx, cury - 1] != room_id
            and curx + 1 < floor.shape[0]
        ):
            curx += 1
            curlen += 1
        if floor[curx, cury] == room_id:
            walls += crawl_rec(curx, cury - 1, room_id, floor, 3, ox, oy)
        else:
            walls += crawl_rec(curx - 1, cury, room_id, floor, 2, ox, oy)
        return walls + [(curx - curlen, cury, curlen, 0)]
    elif dir == 1:
        while (
            floor[curx, cury] == room_id
            and floor[curx, cury + 1] != room_id
            and curx > 0
        ):
            curx -= 1
            curlen += 1
        if floor[curx, cury] == room_id:
            walls += crawl_rec(curx, cury + 1, room_id, floor, 2, ox, oy)
        else:
            walls += crawl_rec(curx + 1, cury, room_id, floor, 3, ox, oy)
        return walls + [(curx + 1, cury + 1, curlen, 0)]
    if dir == 2:
        while (
            floor[curx, cury] == room_id
            and floor[curx + 1, cury] != room_id
            and cury + 1 < floor.shape[1]
        ):  # maybe fix with bounds
            cury += 1
            curlen += 1
        if floor[curx, cury] == room_id:
            walls += crawl_rec(curx + 1, cury, room_id, floor, 0, ox, oy)
        else:
            walls += crawl_rec(curx, cury - 1, room_id, floor, 1, ox, oy)
        return walls + [(curx + 1, cury - curlen, curlen, 1)]
    elif dir == 3:
        while (
            floor[curx, cury] == room_id
            and floor[curx - 1, cury] != room_id
            and cury > 0
        ):
            cury -= 1
            curlen += 1
        if curx == ox and cury + 1 == oy:
            return [(curx, cury + 1, curlen, 1)]
        if floor[curx, cury] == room_id:
            walls += crawl_rec(curx - 1, cury, room_id, floor, 1, ox, oy)
        else:
            walls += crawl_rec(curx, cury + 1, room_id, floor, 0, ox, oy)
        return walls + [(curx, cury + 1, curlen, 1)]


def find_start(room, floor):
    all_room = []
    for i in range(floor.shape[0]):
        for j in range(floor.shape[1]):
            if floor[i, j] == room.room_id:
                all_room.append((i, j))
    minx = min(all_room)[0]
    all_minx = [t for t in all_room if t[0] == minx]
    miny = min(all_minx)[1]
    return minx, miny
