from collections import defaultdict
from typing import Dict, List, Literal, Sequence, Set, Union

import numpy as np
from houses.constants import FLOOR_Y
from houses.utils.types import (
    BoundaryGroups,
    HouseDict,
    ProceduralParameters,
    RoomType,
    Vector3,
    Wall,
    XZPoly,
)
from shapely.geometry import Polygon


def scale_boundary_groups(
    boundary_groups: BoundaryGroups, scale: float, precision: int = 3
) -> BoundaryGroups:
    out = dict()
    for key, lines in boundary_groups.items():
        scaled_lines = set()
        for ((x0, z0), (x1, z1)) in lines:
            scaled_lines.add(
                (
                    (round(x0 * scale, precision), round(z0 * scale, precision)),
                    (round(x1 * scale, precision), round(z1 * scale, precision)),
                )
            )
        out[key] = scaled_lines
    return out


def find_walls(floorplan: np.array):
    walls = defaultdict(list)
    for row in range(len(floorplan) - 1):
        for col in range(len(floorplan[0]) - 1):
            a = floorplan[row, col]
            b = floorplan[row, col + 1]
            if a != b:
                walls[(int(min(a, b)), int(max(a, b)))].append(
                    ((row - 1, col), (row, col))
                )
            b = floorplan[row + 1, col]
            if a != b:
                walls[(int(min(a, b)), int(max(a, b)))].append(
                    ((row, col - 1), (row, col))
                )
    return walls


def get_floor_polygons(xz_poly_map: dict) -> Polygon:
    """Return a shapely Polygon for each floor in the room."""
    floor_polygons = dict()
    for room_id, xz_poly in xz_poly_map.items():
        floor_polygon = []
        for ((x0, z0), (x1, z1)) in xz_poly:
            floor_polygon.append((x0, z0))
        floor_polygon.append((x1, z1))
        floor_polygons[f"room|{room_id}"] = Polygon(floor_polygon)
    return floor_polygons


def get_wall_loop(walls: Sequence[XZPoly]):
    walls_left = set(walls)
    out = [walls[0]]
    walls_left.remove(walls[0])
    while walls_left:
        for wall in walls_left:
            if out[-1][1] == wall[0]:
                out.append(wall)
                walls_left.remove(wall)
                break
            elif out[-1][1] == wall[1]:
                out.append((wall[1], wall[0]))
                walls_left.remove(wall)
                break
        else:
            raise Exception(f"No connecting wall for {out[-1]}!")
    return out


def get_xz_poly_map(boundary_groups, room_ids: Set[int]) -> Dict[int, XZPoly]:
    out = dict()
    for room_id in room_ids:
        room_walls = []
        for k in [k for k in boundary_groups.keys() if room_id in k]:
            room_walls.extend(boundary_groups[k])

        room_wall_loop = get_wall_loop(room_walls)

        # determines if the loop is counter-clockwise, flips if it is
        edge_sum = 0
        for (x0, z0), (x1, z1) in room_wall_loop:
            dist = x0 * z1 - x1 * z0
            edge_sum += dist
        if edge_sum > 0:
            room_wall_loop = [(p1, p0) for p0, p1 in reversed(room_wall_loop)]

        out[room_id] = room_wall_loop
    return out


def create_house(
    xz_poly_map: Dict[int, XZPoly],
    room_type_map: Dict[int, Literal["Bedroom", "Bathroom", "Kitchen", "LivingRoom"]],
    ceiling_height: float,
) -> HouseDict:
    walls = []
    wall_map = dict()
    for room_id, xz_poly in xz_poly_map.items():
        for ((x0, z0), (x1, z1)) in xz_poly:
            wall_id = f"wall|{room_id}|{min(x0, x1):.2f}|{min(z0, z1):.2f}|{max(x0, x1):.2f}|{max(z0, z1):.2f}"
            wall = Wall(
                id=wall_id,
                roomId=f"room|{room_id}",
                polygon=[
                    Vector3(x=x0, y=FLOOR_Y, z=z0),
                    Vector3(x=x1, y=FLOOR_Y, z=z1),
                    Vector3(x=x0, y=FLOOR_Y + ceiling_height, z=z0),
                    Vector3(x=x1, y=FLOOR_Y + ceiling_height, z=z1),
                ],
            )
            walls.append(wall)
            wall_map[wall_id] = wall

    return {
        "wallMap": wall_map,
        "house": HouseDict(
            rooms=[
                RoomType(
                    id=f"room|{room_id}",
                    roomType=room_type_map[room_id],
                    children=[],
                    ceilings=[],
                    floorPolygon=[
                        Vector3(x=x0, y=0, z=z0) for ((x0, z0), (x1, z1)) in xz_poly
                    ],
                )
                for room_id, xz_poly in xz_poly_map.items()
            ],
            walls=walls,
            objects=[],
            proceduralParameters=ProceduralParameters(
                floorColliderThickness=1.0,
                receptacleHeight=0.7,
                lights=[],
                reflections=[],
            ),
        ),
    }
