import gzip
import json
import logging
import random
from collections import Counter
from typing import Dict, Optional

import numpy as np
from ai2thor.controller import Controller
from attrs import define
from houses.constants import PROCTHOR_INITIALIZATION, SCHEMA
from houses.generation.agent import AgentPose, generate_starting_pose
from houses.utils.types import BoundingBox, HouseDict, Vector3
from moviepy.editor import ImageSequenceClip
from shapely.geometry import Point

from .objects import ProceduralRoom
from .room_specs import RoomSpec


@define
class House:
    data: HouseDict
    rooms: Dict[int, ProceduralRoom]
    interior_boundary: np.array
    room_spec: RoomSpec
    add_metadata: bool = True

    def __attrs_post_init__(self) -> None:
        if self.add_metadata:
            self._add_metadata()

    def _add_metadata(self) -> None:
        self.data["metadata"] = {
            "agent": self.choose_agent_pose(),
            "roomSpecId": self.room_spec.room_spec_id,
            "schema": SCHEMA,
        }

    def to_json(self, filename: Optional[str] = None, compressed: bool = False) -> str:
        json_rep = json.dumps(self.data, sort_keys=True, indent=4)
        if filename:
            if compressed:
                assert filename.endswith(".json.gz")
                with gzip.open(filename, "wt", encoding="UTF-8") as f:
                    json.dump(self.data, f)
            else:
                assert filename.endswith(".json")
                with open(filename, "w") as f:
                    f.write(json_rep)
        return json_rep

    @property
    def bounds(self) -> BoundingBox:
        out = BoundingBox(
            min=Vector3(x=float("inf"), y=float("inf"), z=float("inf")),
            max=Vector3(x=float("-inf"), y=float("-inf"), z=float("-inf")),
        )

        for wall in self.data["walls"]:
            for poly in wall["polygon"]:
                for k in ["x", "y", "z"]:
                    if poly[k] < out["min"][k]:
                        out["min"][k] = poly[k]
                    if poly[k] > out["max"][k]:
                        out["max"][k] = poly[k]

        return out

    def choose_agent_pose(self) -> AgentPose:
        """Generate a starting position for the default agent in the house."""
        return generate_starting_pose(self.rooms)

    def validate(self, controller: Controller) -> Dict[str, str]:
        """Validate that the house is useable.

        Returns a dictionary of warnings or errors.
        """
        warnings = {}
        controller.reset(renderImage=False)
        event = controller.step(
            action="CreateHouse", house=self.data, renderImage=False
        )
        if not event:
            warnings["CreateHouse"] = "Failed to create house."
            logging.warning(warnings)
            self.data["metadata"]["warnings"] = warnings
            return warnings
        event = controller.step(
            action="TeleportFull",
            **self.data["metadata"]["agent"],
            renderImage=False,
        )
        if not event:
            warnings["TeleportFull"] = "Unable to teleport to starting position."
            logging.warning(warnings)
            self.data["metadata"]["warnings"] = warnings
            return warnings
        event = controller.step(action="GetReachablePositions", renderImage=False)
        if not event:
            warnings["GetReachablePositions"] = "Failed to get reachable positions"
            logging.warning(warnings)
            self.data["metadata"]["warnings"] = warnings
            return warnings
        rps = event.metadata["actionReturn"]

        random.shuffle(rps)

        # Check that every room is reachable
        points_per_room = Counter({r: 0 for r in self.rooms})
        for p in rps:
            point = Point(p["x"], p["z"])
            for room_id in self.rooms.keys():
                room_poly = self.rooms[room_id].room_polygon.polygon
                if room_poly.contains(point):
                    points_per_room[room_id] += 1
                    break
            else:
                warnings[f"Unreachable point: {p}"] = "Unreachable point."

            if all(num_points >= 5 for num_points in points_per_room.values()):
                break
        else:
            unnavigable_rooms = list(
                filter(lambda r: points_per_room[r] < 5, points_per_room)
            )
            warnings[
                "RoomsNotNavigable"
            ] = f"Rooms {unnavigable_rooms} / {len(points_per_room)} are not navigable."

        if warnings:
            logging.warning(warnings)

        self.data["metadata"]["warnings"] = warnings

        return warnings

    @staticmethod
    def top_down_video(
        data: HouseDict,
        controller: Controller,
        filename: str,
        width: int = 1024,
        height: int = 1024,
    ) -> None:
        """Saves a top-down video of the house."""
        if controller is None:
            controller = Controller(
                width=width, height=height, **PROCTHOR_INITIALIZATION
            )

        controller.step(action="CreateHouse", house=data)
        max_meters = max(
            max(p["x"], p["z"]) for wall in data["walls"] for p in wall["polygon"]
        )
        raise NotImplementedError("Update max_meters from data.")

        if not controller.last_event.third_party_camera_frames:
            controller.step(
                action="AddThirdPartyCamera",
                position=dict(x=0, y=0, z=0),
                rotation=dict(x=0, y=0, z=0),
                skyboxColor="white",
            )

        dist = 2.75
        images = []
        for angle in range(0, 360, 45):
            controller.step(
                action="UpdateThirdPartyCamera",
                position=dict(
                    x=max_meters / 2 - np.sin(angle * np.pi / 180) * dist,
                    y=max_meters + 1.5,
                    z=max_meters / 2 - np.cos(angle * np.pi / 180) * dist,
                ),
                rotation=dict(x=65, y=angle, z=0),
            )
            images.append(controller.last_event.third_party_camera_frames[0])
        imsn = ImageSequenceClip(images, fps=2)
        imsn.write_videofile(filename)
